#! /usr/bin/env python3

from itertools import chain
import platform
import os
import subprocess

from lab.environments import LocalEnvironment, BaselSlurmEnvironment
from lab import tools

from downward.experiment import FastDownwardExperiment
from downward.reports.absolute import AbsoluteReport
from downward.reports.scatter import ScatterPlotReport

def get_script():
    """Get file name of main script."""
    return tools.get_script_path()


def get_script_dir():
    """Get directory of main script.

    Usually a relative directory (depends on how it was called by the user.)"""
    return os.path.dirname(get_script())

def get_repo_base():
    """Get base directory of the repository, as an absolute path.

    Search upwards in the directory tree from the main script until a
    directory with a subdirectory named ".hg" is found.

    Abort if the repo base cannot be found."""
    path = os.path.abspath(get_script_dir())
    while os.path.dirname(path) != path:
        if os.path.exists(os.path.join(path, ".hg")):
            return path
        path = os.path.dirname(path)
    sys.exit("repo base could not be found")

def is_running_on_cluster():
    node = platform.node()
    return node.endswith(".scicore.unibas.ch") or node.endswith(".cluster.bc2.ch")

def get_algo_nick(revision, config_nick):
    return "{revision}-{config_nick}".format(**locals())

DIR = os.path.dirname(os.path.abspath(__file__))
SCRIPT_NAME = os.path.splitext(os.path.basename(__file__))[0]
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
SUITE = [ # optimal strips
    'agricola-opt18-strips', 'airport', 'barman-opt11-strips',
    'barman-opt14-strips', 'blocks', 'childsnack-opt14-strips',
    'data-network-opt18-strips', 'depot', 'driverlog',
    'elevators-opt08-strips', 'elevators-opt11-strips',
    'floortile-opt11-strips', 'floortile-opt14-strips', 'freecell',
    'ged-opt14-strips', 'grid', 'gripper', 'hiking-opt14-strips',
    'logistics00', 'logistics98', 'miconic', 'movie', 'mprime',
    'mystery', 'nomystery-opt11-strips', 'openstacks-opt08-strips',
    'openstacks-opt11-strips', 'openstacks-opt14-strips',
    'openstacks-strips', 'organic-synthesis-opt18-strips',
    'organic-synthesis-split-opt18-strips', 'parcprinter-08-strips',
    'parcprinter-opt11-strips', 'parking-opt11-strips',
    'parking-opt14-strips', 'pathways-noneg', 'pegsol-08-strips',
    'pegsol-opt11-strips', 'petri-net-alignment-opt18-strips',
    'pipesworld-notankage', 'pipesworld-tankage', 'psr-small', 'rovers',
    'satellite', 'scanalyzer-08-strips', 'scanalyzer-opt11-strips',
    'snake-opt18-strips', 'sokoban-opt08-strips',
    'sokoban-opt11-strips', 'spider-opt18-strips', 'storage',
    'termes-opt18-strips', 'tetris-opt14-strips',
    'tidybot-opt11-strips', 'tidybot-opt14-strips', 'tpp',
    'transport-opt08-strips', 'transport-opt11-strips',
    'transport-opt14-strips', 'trucks-strips', 'visitall-opt11-strips',
    'visitall-opt14-strips', 'woodworking-opt08-strips',
    'woodworking-opt11-strips', 'zenotravel']
DEFAULT_TABLE_ATTRIBUTES = [
    "cost",
    "coverage",
    "error",
    "evaluations",
    "expansions",
    "expansions_until_last_jump",
    "generated",
    "memory",
    "planner_memory",
    "planner_time",
    "quality",
    "run_dir",
    "score_evaluations",
    "score_expansions",
    "score_generated",
    "score_memory",
    "score_search_time",
    "score_total_time",
    "search_time",
    "total_time",
    ]
ENVIRONMENT = BaselSlurmEnvironment(
    partition="infai_2",
    email="silvan.sievers@unibas.ch",
    export=["PATH", "DOWNWARD_BENCHMARKS"])

if not is_running_on_cluster():
    SUITE = ["depot:p01.pddl", "gripper:prob01.pddl"]
    ENVIRONMENT = LocalEnvironment(processes=4)

class IssueConfig(object):
    """Hold information about a planner configuration.

    See FastDownwardExperiment.add_algorithm() for documentation of the
    constructor's options.

    """
    def __init__(self, nick, component_options,
                 build_options=None, driver_options=None):
        self.nick = nick
        self.component_options = component_options
        self.build_options = build_options
        self.driver_options = driver_options

REVISION = "bbe8886982ca"
CONFIGS = list(chain.from_iterable(
    (
    IssueConfig(
        "{heuristic}-{pruning}-full-min-{min_ratio}".format(**locals()),
        ["--search", "astar({heuristic}(), pruning=stubborn_sets_{pruning}(only_active_operators=false,mutex_interference=false,use_wss=false,min_required_pruning_ratio={min_ratio}))".format(**locals())]),
    IssueConfig(
        "{heuristic}-{pruning}-full-mutex-min-{min_ratio}".format(**locals()),
        ["--search", "astar({heuristic}(), pruning=stubborn_sets_{pruning}(only_active_operators=false,mutex_interference=true,use_wss=false,min_required_pruning_ratio={min_ratio}))".format(**locals())]),
    )
    for heuristic in ["blind", "lmcut"]
    for pruning in ["ec"]
    for min_ratio in [0.0] # to turn on, use 0.2
))

exp = FastDownwardExperiment(
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)
exp.add_parser(exp.EXITCODE_PARSER)
exp.add_parser(exp.TRANSLATOR_PARSER)
exp.add_parser(exp.SINGLE_SEARCH_PARSER)
exp.add_parser(exp.PLANNER_PARSER)
exp.add_parser("ss-parser.py")

for config in CONFIGS:
    exp.add_algorithm(
        get_algo_nick(REVISION, config.nick),
        get_repo_base(),
        REVISION,
        config.component_options,
        build_options=config.build_options,
        driver_options=config.driver_options)


exp.add_step('build', exp.build)
exp.add_step('start', exp.start_runs)
exp.add_fetcher(name='fetch')

attributes = list(DEFAULT_TABLE_ATTRIBUTES)
attributes.extend([
    "ss_successors_unpruned",
    "ss_successors_pruned",
    "ss_pruning_ratio",
    "ss_pruning_time",
])
outfile=os.path.join(exp.eval_dir, exp.name + '-abs.html')
exp.add_report(
    AbsoluteReport(
        attributes=attributes,
    ),
    name='absolute',
    outfile=outfile,
)
exp.add_step(
    'publish-absolute-report', subprocess.call, ['publish', outfile])

exp.run_steps()
