#! /usr/bin/env python3

import platform
import os

from lab.environments import LocalEnvironment, BaselSlurmEnvironment
from lab import tools

from downward.experiment import FastDownwardExperiment
from downward.reports.compare import ComparativeReport
from downward.reports.scatter import ScatterPlotReport

def get_script():
    """Get file name of main script."""
    return tools.get_script_path()


def get_script_dir():
    """Get directory of main script.

    Usually a relative directory (depends on how it was called by the user.)"""
    return os.path.dirname(get_script())

def get_repo_base():
    """Get base directory of the repository, as an absolute path.

    Search upwards in the directory tree from the main script until a
    directory with a subdirectory named ".hg" is found.

    Abort if the repo base cannot be found."""
    path = os.path.abspath(get_script_dir())
    while os.path.dirname(path) != path:
        if os.path.exists(os.path.join(path, ".hg")):
            return path
        path = os.path.dirname(path)
    sys.exit("repo base could not be found")

def is_running_on_cluster():
    node = platform.node()
    return node.endswith(".scicore.unibas.ch") or node.endswith(".cluster.bc2.ch")

def get_algo_nick(revision, config_nick):
    return "{revision}-{config_nick}".format(**locals())

DIR = os.path.dirname(os.path.abspath(__file__))
SCRIPT_NAME = os.path.splitext(os.path.basename(__file__))[0]
BENCHMARKS_DIR = os.environ["DOWNWARD_BENCHMARKS"]
SUITE = [ # optimal strips
    'agricola-opt18-strips', 'airport', 'barman-opt11-strips',
    'barman-opt14-strips', 'blocks', 'childsnack-opt14-strips',
    'data-network-opt18-strips', 'depot', 'driverlog',
    'elevators-opt08-strips', 'elevators-opt11-strips',
    'floortile-opt11-strips', 'floortile-opt14-strips', 'freecell',
    'ged-opt14-strips', 'grid', 'gripper', 'hiking-opt14-strips',
    'logistics00', 'logistics98', 'miconic', 'movie', 'mprime',
    'mystery', 'nomystery-opt11-strips', 'openstacks-opt08-strips',
    'openstacks-opt11-strips', 'openstacks-opt14-strips',
    'openstacks-strips', 'organic-synthesis-opt18-strips',
    'organic-synthesis-split-opt18-strips', 'parcprinter-08-strips',
    'parcprinter-opt11-strips', 'parking-opt11-strips',
    'parking-opt14-strips', 'pathways-noneg', 'pegsol-08-strips',
    'pegsol-opt11-strips', 'petri-net-alignment-opt18-strips',
    'pipesworld-notankage', 'pipesworld-tankage', 'psr-small', 'rovers',
    'satellite', 'scanalyzer-08-strips', 'scanalyzer-opt11-strips',
    'snake-opt18-strips', 'sokoban-opt08-strips',
    'sokoban-opt11-strips', 'spider-opt18-strips', 'storage',
    'termes-opt18-strips', 'tetris-opt14-strips',
    'tidybot-opt11-strips', 'tidybot-opt14-strips', 'tpp',
    'transport-opt08-strips', 'transport-opt11-strips',
    'transport-opt14-strips', 'trucks-strips', 'visitall-opt11-strips',
    'visitall-opt14-strips', 'woodworking-opt08-strips',
    'woodworking-opt11-strips', 'zenotravel']
ENVIRONMENT = BaselSlurmEnvironment(
    partition="infai_2",
    email="silvan.sievers@unibas.ch",
    export=["PATH", "DOWNWARD_BENCHMARKS"])

if not is_running_on_cluster():
    SUITE = ["depot:p01.pddl", "gripper:prob01.pddl"]
    ENVIRONMENT = LocalEnvironment(processes=4)

class IssueConfig(object):
    """Hold information about a planner configuration.

    See FastDownwardExperiment.add_algorithm() for documentation of the
    constructor's options.

    """
    def __init__(self, nick, component_options,
                 build_options=None, driver_options=None):
        self.nick = nick
        self.component_options = component_options
        self.build_options = build_options
        self.driver_options = driver_options

REVISION = "bbe8886982ca"
CONFIGS = [
    IssueConfig("translate", [], None, driver_options=['--translate']),
]

exp = FastDownwardExperiment(
    environment=ENVIRONMENT,
)
exp.add_suite(BENCHMARKS_DIR, SUITE)
exp.add_parser(exp.EXITCODE_PARSER)
exp.add_parser(exp.TRANSLATOR_PARSER)

for config in CONFIGS:
    exp.add_algorithm(
        get_algo_nick(REVISION, config.nick),
        get_repo_base(),
        REVISION,
        config.component_options,
        build_options=config.build_options,
        driver_options=config.driver_options)


exp.add_step('build', exp.build)
exp.add_step('start', exp.start_runs)
exp.add_fetcher(name='fetch')

def invent_operators_algo(run):
    if run['algorithm'] == '{}-translate'.format(REVISION):
        new_algorithm = 'translate-operators'
        new_run = {}
        new_run['id'] = [new_algorithm, run['domain'], run['problem']]
        new_run['algorithm'] = new_algorithm
        new_run['domain'] = run['domain']
        new_run['problem'] = run['problem']
        new_run['the_attribute'] = run.get('translator_operators')
        return new_run
    return True

def invent_propositions_algo(run):
    if run['algorithm'] == '{}-translate'.format(REVISION):
        new_algorithm = 'translate-propositions'
        new_run = {}
        new_run['id'] = [new_algorithm, run['domain'], run['problem']]
        new_run['algorithm'] = new_algorithm
        new_run['domain'] = run['domain']
        new_run['problem'] = run['problem']
        new_run['the_attribute'] = run.get('translator_facts')
        return new_run
    return True

exp.add_fetcher('data/2020-02-08-B-number-propositions-vs-operators-eval', name='fetch-copy-operators', filter=[invent_operators_algo])
exp.add_fetcher('data/2020-02-08-B-number-propositions-vs-operators-eval', name='fetch-copy-propositions', filter=[invent_propositions_algo])

exp.add_report(
    ComparativeReport(
        algorithm_pairs=[
            ('translate-operators', 'translate-propositions'),
        ],
        attributes=['the_attribute'],
    ),
    name='compare-propositions-vs-operators',
)

recognized_as_unsolvable = [
    ('mystery', 'prob07.pddl'),
    ('mystery', 'prob18.pddl'),
]
def filter_unsolvable_tasks(run):
    return (run['domain'], run['problem']) not in recognized_as_unsolvable

domain_groups_optimal = {
'agricola-opt18-strips': 'agricola',
'airport': 'airport', # identity
'assembly': 'assembly', # identity
'barman-opt11-strips': 'barman',
'barman-opt14-strips': 'barman',
'blocks': 'blocks', # identity
'caldera-opt18-adl': 'caldera',
'caldera-split-opt18-adl': 'caldera-split',
'cavediving-14-adl': 'cavediving',
'childsnack-opt14-strips': 'childsnack',
'citycar-opt14-adl': 'citycar',
'data-network-opt18-strips': 'data-network',
'depot': 'depot', # identity
'driverlog': 'driverlog', # identity
'elevators-opt08-strips': 'elevators',
'elevators-opt11-strips': 'elevators',
'floortile-opt11-strips': 'floortile',
'floortile-opt14-strips': 'floortile',
'freecell': 'freecell', # identity
'ged-opt14-strips': 'ged',
'grid': 'grid', # identity
'gripper': 'gripper', # identity
'hiking-opt14-strips': 'hiking',
'logistics00': 'logistics',
'logistics98': 'logistics',
'maintenance-opt14-adl': 'maintenance',
'miconic': 'miconic', # identity
'miconic-fulladl': 'miconic',
'miconic-simpleadl': 'miconic',
'movie': 'movie', # identity
'mprime': 'mprime', # identity
'mystery': 'mystery', # identity
'nomystery-opt11-strips': 'nomystery',
'nurikabe-opt18-adl': 'nurikabe',
'openstacks': 'openstacks', # identity
'openstacks-opt08-adl': 'openstacks',
'openstacks-opt08-strips': 'openstacks',
'openstacks-opt11-strips': 'openstacks',
'openstacks-opt14-strips': 'openstacks',
'openstacks-strips': 'openstacks',
'optical-telegraphs': 'optical-telegraphs', # identity
'organic-synthesis-opt18-strips': 'organic-synthesis',
'organic-synthesis-split-opt18-strips': 'organic-synthesis-split',
'parcprinter-08-strips': 'parcprinter',
'parcprinter-opt11-strips': 'parcprinter',
'parking-opt11-strips': 'parking',
'parking-opt14-strips': 'parking',
'pathways': 'pathways', # identity
'pathways-noneg': 'pathways',
'pegsol-08-strips': 'pegsol',
'pegsol-opt11-strips': 'pegsol',
'petri-net-alignment-opt18-strips': 'petri-net-alignment',
'philosophers': 'philosophers', # identity
'pipesworld-notankage': 'pipesworld-notankage', # pipesworld?
'pipesworld-tankage': 'pipesworld-tankage', # pipesworld?
'psr-large': 'psr',
'psr-middle': 'psr',
'psr-small': 'psr',
'rovers': 'rovers', # identity
'satellite': 'satellite', # identity
'scanalyzer-08-strips': 'scanalyzer',
'scanalyzer-opt11-strips': 'scanalyzer',
'schedule': 'schedule', # identity
'settlers-opt18-adl': 'settlers',
'snake-opt18-strips': 'snake',
'sokoban-opt08-strips': 'sokoban',
'sokoban-opt11-strips': 'sokoban',
'spider-opt18-strips': 'spider',
'storage': 'storage', # identity
'termes-opt18-strips': 'termes',
'tetris-opt14-strips': 'tetris',
'tidybot-opt11-strips': 'tidybot',
'tidybot-opt14-strips': 'tidybot',
'tpp': 'tpp', # identity
'transport-opt08-strips': 'transport',
'transport-opt11-strips': 'transport',
'transport-opt14-strips': 'transport',
'trucks': 'trucks', # identity
'trucks-strips': 'trucks',
'visitall-opt11-strips': 'visitall',
'visitall-opt14-strips': 'visitall',
'woodworking-opt08-strips': 'woodworking',
'woodworking-opt11-strips': 'woodworking',
'zenotravel': 'zenotravel', # identity
}

def domain_as_category(run1, run2):
    return domain_groups_optimal[run1['domain']]

interesting_domains = [
'woodworking',
'parcprinter',
'airport',
'openstacks',
'tpp',
]

def interesting_domains_as_category(run1, run2):
    grouped_domain = domain_groups_optimal[run1['domain']]
    if grouped_domain in interesting_domains:
        return grouped_domain
    return ''

exp.add_report(
    ScatterPlotReport(
        filter=[filter_unsolvable_tasks],
        filter_algorithm=['translate-operators', 'translate-propositions'],
        attributes=['the_attribute'],
        format='tex',
        scale='log',
        show_missing=False,
        # get_category=domain_as_category,
    ),
    name='scatter-propositions-vs-operators',
)

exp.run_steps()
